DrawingBoard = LibStub("AceAddon-3.0"):NewAddon("DrawingBoard", "AceConsole-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("DrawingBoard")


DrawingBoard.debug = true


function DrawingBoard:OnInitialize()
    -- Called when the addon is loaded
   
end



function DrawingBoard:OnEnable()
    -- Called when the addon is enabled
    local defaults = { profile = {
				    bc_title = L["New Board"],
				    bc_prefix = L["Assignment "],
				    bc_channel = "raid"
				}
		    }
    self.db = LibStub("AceDB-3.0"):New("DrawingBoardDB", defaults, "Default")
    
    
    
    self.frames = {}
    self.boards = {}
    self.activeboard = {}
    
    self:CreateMainFrames()
    self:NewBoard()
    self:Close()
    
    self:RegisterChatCommand("db", "Open")
end



function DrawingBoard:OnDisable()
    -- Called when the addon is disabled
end




-- ********* Window functions

function DrawingBoard:Open()      
    for name, frame in pairs(self.frames) do
	frame:Show()
    end
    for name, frame in pairs(self.activeboard.frames) do
	frame:Show()
    end
end


function DrawingBoard:Close()
    for name, frame in pairs(self.frames) do
	frame:Hide()
    end
    for name, frame in pairs(self.activeboard.frames) do
	frame:Hide()
    end
end

function DrawingBoard:DeleteBoard()
    self.frames["DrawingBoardBoard"..self.activeboard.number.."Button"]:Hide()
    self.frames["DrawingBoardBoard"..self.activeboard.number.."Button"] = nil
    
    for name, frame in pairs(self.activeboard.frames) do
	frame:Hide()
	self.activeboard.frames[name] = nil
    end
    
    local i = 0
    for number, board in pairs(self.boards) do
	if self.activeboard.number ~= number then i = number end
    end
    self.boards[self.activeboard.number] = nil
    if i == 0 then
	self:NewBoard()
    else
	self:ShowBoard(i)
	self:SortBoardButtons(self.activeboard)
    end
    
end


function DrawingBoard:NewBoard()
    -- Chose new board number
    number = 0
    for board_num,_  in pairs(self.boards) do
	number = math.max(number, board_num)
    end
    number = number + 1
    
    -- Create Empty Board
    local b = {}
    b.prefix = self.db.profile.bc_prefix
    b.channel = self.db.profile.bc_channel
    b.number = number
    b.frames = {}
    self.boards[number] = b
    
    
    -- Create Board Button
    local b = CreateFrame("Button", "DrawingBoardBoard"..number.."Button", f, "OptionsButtonTemplate")
    b:SetFrameStrata("FULLSCREEN")
    b.number = number
    b:SetScript("OnClick", function () DrawingBoard:ShowBoard(b.number) end)
    self.frames["DrawingBoardBoard"..number.."Button"] = b
    
    self:RenameBoard(self.boards[number], self.db.profile.bc_title)
    self:ShowBoard(number)
    self:SortBoardButtons(self.activeboard)
    
    
    if GetNumRaidMembers() > 0 then
	DrawingBoard:AddPlayers()
    end
    for i=1,8 do
	DrawingBoard:CreateRTFrame(i)
    end
end


function DrawingBoard:AddPlayers()
    local classlist = {}
    local class_names = {}
    local class
    local name
    for i=1,GetNumRaidMembers() do
	_, class = UnitClass("raid"..i)
	name = UnitName("raid"..i)
	if not class_names[class] then
	    class_names[class] = {}
	    table.insert(classlist, class)
	end
	table.insert(class_names[class], name)
    end
    
    table.sort(classlist)
    for _, names in pairs(class_names) do
	table.sort(names)
    end
    
    local i = 1
    for _, class in ipairs(classlist) do
	for _, name in ipairs(class_names[class]) do
	    DrawingBoard:CreatePlayerFrame(name, class, i)
	    i = i + 1
	end
    end
    
end


function DrawingBoard:ShowBoard(boardnumber)
    for name, frame in pairs(self.activeboard.frames or {}) do
	frame:Hide()
    end
    
    for name, frame in pairs(self.boards[boardnumber].frames) do
	frame:Show()
    end
    
    self.activeboard = self.boards[boardnumber]
    self.frames["DrawingBoardMainFrame"].title:SetText(self.activeboard.title)
end


function DrawingBoard:RenameBoard(board, newname)
    board.title = newname
    
    local b = self.frames["DrawingBoardBoard"..board.number.."Button"]
    b:SetText(newname)
    b:SetWidth(b:GetFontString():GetStringWidth()+16)
end



function DrawingBoard:SortBoardButtons()
    y = 0
    lastnumber = -1
    for number, board in pairs(self.boards) do
	self.frames["DrawingBoardBoard"..number.."Button"]:ClearAllPoints()
	self.frames["DrawingBoardBoard"..number.."Button"]:SetPoint("TOPRIGHT", self.frames["DrawingBoardMainFrame"], "TOPLEFT", -2, -y)
	y= y + self.frames["DrawingBoardBoard"..number.."Button"]:GetHeight() + 2
	lastnumber = number
    end
    self.frames["DrawingBoardNewBoardButton"]:ClearAllPoints()
    if lastnumber ~= -1 then
	self.frames["DrawingBoardNewBoardButton"]:SetPoint("TOPRIGHT", self.frames["DrawingBoardBoard"..lastnumber.."Button"], "BOTTOMRIGHT", 0, -2)
    else
	self.frames["DrawingBoardNewBoardButton"]:SetPoint("TOPRIGHT", self.frames["DrawingBoardMainFrame"], "TOPLEFT", -2, 0)
    end
end

function DrawingBoard:Broadcast(board)
    local message
    local first
    local assignments = {}
  -- ToDo : Ausgabe sortieren.
    self:SendMessage(board.title, board.channel)
    for name, frame in pairs(board.frames) do
	if frame.is_assignment then
	    message = board.prefix ..frame:GetText() ..": "
	    first = true
	    for name2, frame2 in pairs(frame.attached) do
		if not first then message = message .. ", " end
		first = false
		message = message .. frame2:GetText()
	    end
	    self:SendMessage(message, board.channel)
	end
    end
end


-- ********** Create Frames

function DrawingBoard:CreateMainFrames()
    local f = CreateFrame("Frame", "DrawingBoardMainFrame", UIParent)
    f:SetFrameStrata("FULLSCREEN")
    f:SetWidth(700)
    f:SetHeight(500)
    f:SetBackdrop({bgFile = "Interface/Tooltips/UI-Tooltip-Background", 
                                            edgeFile = "Interface/Tooltips/UI-Tooltip-Border", 
                                            tile = true, tileSize = 32, edgeSize = 8, 
                                            insets = { left = 0, right = 0, top = 0, bottom = 0 }})
    f:SetBackdropColor(0, 0, 0, 0.8)
    f:SetClampedToScreen(true)
    f:SetPoint("TOPLEFT", UIParent, "TOPLEFT", 150, -120)
    
    f.title = f:CreateFontString(nil, "ARTWORK")
    f.title:SetFontObject(GameFontNormalLarge)
    f.title:SetJustifyH("LEFT")
    f.title:SetPoint("TOPLEFT",6,-6)
    
    self.frames["DrawingBoardMainFrame"] = f
    
    local b = CreateFrame("Button", "DrawingBoardCustomButton", f, "OptionsButtonTemplate")
    b:SetFrameStrata("FULLSCREEN")
    b:SetPoint("TOPRIGHT", f, "BOTTOMRIGHT", 0, -2)
    b:SetText(L["Custom Assignment"])
    b:SetWidth(b:GetFontString():GetStringWidth()+16)
    b:SetScript("OnClick", function () DrawingBoard:ShowCustomDialog() end)
    self.frames["DrawingBoardCustomButton"] = b
    
    local b = CreateFrame("Button", "DrawingBoardCloseButton", f, "OptionsButtonTemplate")
    b:SetFrameStrata("FULLSCREEN")
    b:SetPoint("BOTTOMRIGHT", f, "TOPRIGHT", 0, 2)
    b:SetText("X")
    b:SetWidth(b:GetHeight()+10)
    b:SetScript("OnClick", function () DrawingBoard:Close() end)
    self.frames["DrawingBoardCloseButton"] = b
    
    local b = CreateFrame("Button", "DrawingBoardDeleteBoardButton", f, "OptionsButtonTemplate")
    b:SetFrameStrata("FULLSCREEN")
    b:SetPoint("RIGHT", self.frames["DrawingBoardCloseButton"], "LEFT", -2, 0)
    b:SetText(L["Delete Board"])
    b:SetWidth(b:GetFontString():GetStringWidth()+16)
    b:SetScript("OnClick", function () DrawingBoard:DeleteBoard() end)
    self.frames["DrawingBoardDeleteBoardButton"] = b
    
    
    local b = CreateFrame("Button", "DrawingBoardBoardOptionsButton", f, "OptionsButtonTemplate")
    b:SetFrameStrata("FULLSCREEN")
    b:SetPoint("BOTTOMLEFT", self.frames["DrawingBoardMainFrame"], "TOPLEFT", 0, 2)
    b:SetText(L["Board Options"])
    b:SetWidth(b:GetFontString():GetStringWidth()+16)
    b:SetScript("OnClick", function () DrawingBoard:ShowBoardOptionsDialog() end)
    self.frames["DrawingBoardBoardOptionsButton"] = b
    
    local b = CreateFrame("Button", "DrawingBoardNewBoardButton", f, "OptionsButtonTemplate")
    b:SetFrameStrata("FULLSCREEN")
    b:SetText("+")
    b:SetWidth(b:GetHeight()+10)
    b:SetScript("OnClick", function () DrawingBoard:NewBoard() end)
    self.frames["DrawingBoardNewBoardButton"] = b
    
    local b = CreateFrame("Button", "DrawingBoardBroadcastButton", f, "OptionsButtonTemplate")
    b:SetFrameStrata("FULLSCREEN")
    b:SetPoint("LEFT", self.frames["DrawingBoardBoardOptionsButton"], "RIGHT", 2, 0)
    b:SetText(L["Broadcast Board"])
    b:SetWidth(b:GetFontString():GetStringWidth()+16)
    b:SetScript("OnClick", function () DrawingBoard:Broadcast(DrawingBoard.activeboard) end)
    self.frames["DrawingBoardBroadcastButton"] = b
    
    
    
  --[[ local b = CreateFrame("Button", "DrawingBoardBroadcastOptionsButton", f, "OptionsButtonTemplate")
    b:SetFrameStrata("FULLSCREEN")
    b:SetPoint("BOTTOMRIGHT", self.frames["DrawingBoardBroadcastButton"], "TOPRIGHT", 0, 2)
    b:SetText("Broadcast Options")
    b:SetWidth(b:GetFontString():GetStringWidth()+10)
    b:SetScript("OnClick", function () DrawingBoard:ShowBroadcastOptionsDialog() end)
    self.frames["DrawingBoardBroadcastOptionsButton"] = b --]]
end



function DrawingBoard:ShowBoardOptionsDialog()
    local d = LibStub("AceGUI-3.0"):Create("Frame")
    local title = LibStub("AceGUI-3.0"):Create("EditBox")
    local prefix = LibStub("AceGUI-3.0"):Create("EditBox")
    local channel = LibStub("AceGUI-3.0"):Create("EditBox")
    local default = LibStub("AceGUI-3.0"):Create("Button")

    
    d:SetTitle(L["Board Options"])
    d:SetWidth(400)
    d:SetHeight(225)
    
    title:SetText(DrawingBoard.activeboard.title)
    title:SetLabel(L["Title"])
    title:SetCallback("OnEnterPressed", function(self, event, text)
					DrawingBoard.activeboard.title = text
					DrawingBoard.frames["DrawingBoardBoard"..DrawingBoard.activeboard.number.."Button"]:SetText(text)
					DrawingBoard.frames["DrawingBoardBoard"..DrawingBoard.activeboard.number.."Button"]:SetWidth(DrawingBoard.frames["DrawingBoardBoard"..DrawingBoard.activeboard.number.."Button"]:GetFontString():GetWidth()+16)
					DrawingBoard.frames["DrawingBoardMainFrame"].title:SetText(text)
				    end)
				    
    prefix:SetText(DrawingBoard.activeboard.prefix)
    prefix:SetLabel(L["Prefix"])
    prefix:SetCallback("OnEnterPressed", function(self, event, text)
					DrawingBoard.activeboard.prefix = text
				    end)
    
    channel:SetText(DrawingBoard.activeboard.channel)
    channel:SetLabel(L["Channel"])
    channel:SetCallback("OnEnterPressed", function(self, event, text)
					DrawingBoard.activeboard.channel = text
				    end)
				    
    default:SetText(L["Set to default"])
    default:SetCallback("OnClick", function (self, event)
					DrawingBoard.db.profile.bc_title = DrawingBoard.activeboard.title
					DrawingBoard.db.profile.bc_prefix = DrawingBoard.activeboard.prefix
					DrawingBoard.db.profile.bc_channel = DrawingBoard.activeboard.channel
				    end)
        	    
    d:AddChild(title)
    d:AddChild(prefix)
    d:AddChild(channel)
    d:AddChild(default)
    d:Show()
end


function DrawingBoard:ShowCustomDialog()
    local d = LibStub("AceGUI-3.0"):Create("Frame")
    local e = LibStub("AceGUI-3.0"):Create("EditBox")
    
    d:SetWidth(400)
    d:SetHeight(200)
    
    d:SetTitle(L["Custom Assignment"])
    e:SetCallback("OnEnterPressed", function(self, event, text)
					DrawingBoard:CreateCustomFrame(text)
					d:SetStatusText("Custom Button "..text.." created")
					e:SetText("")
				    end)

    d:AddChild(e)
    d:Show()
    e.editbox:SetFocus()
end


function DrawingBoard:CreateFrame(f, text)
    f:SetFrameStrata("FULLSCREEN")
    f:SetBackdrop({bgFile = "Interface/Tooltips/UI-Tooltip-Background", 
                                            edgeFile = "Interface/Tooltips/UI-Tooltip-Border", 
                                            tile = true, tileSize = 16, edgeSize = 8, 
                                            insets = { left = 0, right = 0, top = 0, bottom = 0 }})
    f:SetBackdropColor(0, 0, 0, 1)
    f:SetClampedToScreen(true)
    
    f.text = f:CreateFontString(nil, "ARTWORK")
    f.text:SetFontObject(GameFontNormal)
    f.text:SetJustifyH("LEFT")
    f.text:SetPoint("LEFT",6,0)
    f.text:SetText(text)
    
    f:SetWidth(f.text:GetStringWidth()+12)
    f:SetHeight(f.text:GetStringHeight()+7)
    
    f:SetMovable(true)
    f:RegisterForDrag("LeftButton")
    f:EnableMouse(true)
    f:SetScript("OnDragStart", f.StartMoving)
    f:SetScript("OnDragStop", function () f:StopMovingOrSizing(); self:DropButton(f) end )
    
    f.attached = {}
end


function DrawingBoard:CreateRTFrame(number)
    local f = CreateFrame("Frame", "DrawingBoardBoard"..self.activeboard.number.."RT"..number.."Frame", UIParent)
    self:CreateFrame(f, "")
    f:SetWidth(25)
    f:SetHeight(25)
    f.GetText = function (self) return "{rt"..number.."}" end
    
    f.texture = f:CreateTexture(nil, "OVERLAY")
    f.texture:SetTexture("Interface\\TargetingFrame\\UI-RaidTargetingIcons")
    f.texture:SetPoint("TOPLEFT", f, "TOPLEFT", 2, -2)
    f.texture:SetPoint("BOTTOMRIGHT", f, "BOTTOMRIGHT", -2, 2)
    SetRaidTargetIconTexture(f.texture,number)
    
    f:SetPoint("TOPLEFT", self.frames["DrawingBoardMainFrame"], "BOTTOMLEFT", (number-1)*(f:GetWidth()+2), -2)
    
    f.type = "rt"
    f.number = number
    self.activeboard.frames["DrawingBoardBoard"..self.activeboard.number.."RT"..number.."Frame"] = f
end

function DrawingBoard:CreatePlayerFrame(name, class, number)
    local f = CreateFrame("Frame", "DrawingBoardBoard"..self.activeboard.number.."Player"..number.."Frame", UIParent)
    self:CreateFrame(f,name)
    
    f:SetPoint("TOPLEFT", self.frames["DrawingBoardMainFrame"], "TOPLEFT", 705, -20*(number-1))
    f.text:SetTextColor(RAID_CLASS_COLORS[class].r,RAID_CLASS_COLORS[class].g,RAID_CLASS_COLORS[class].b,1)
    
    f.GetText = function (self) return f.text:GetText() end
    
    f.type = "player"
    f.class = class
    f.number = number
    self.activeboard.frames["DrawingBoardBoard"..self.activeboard.number.."Player"..number.."Frame"] = f
end


function DrawingBoard:CreateCustomFrame(text)
    local number = 0
    repeat
	number = random(100000)
    until not self.frames["DrawingBoardCustom"..number.."Frame"]
    
    local f = CreateFrame("Frame", "DrawingBoardBoard"..self.activeboard.number.."Custom"..number.."Frame", UIParent)
    self:CreateFrame(f,text)
    
    f.text:SetTextColor(1,1,1,1)
    f:SetPoint("TOPLEFT", self.frames["DrawingBoardMainFrame"], "TOPLEFT", 350, -505)
    
    f.GetText = function (self) return f.text:GetText() end
    
    f.type = "custom"
    self.activeboard.frames["DrawingBoardBoard"..self.activeboard.number.."Custom"..number.."Frame"] = f
end



-- ******* Button functions


function DrawingBoard:ResetButton(button, deleting)
    local self = DrawingBoard
    button:ClearAllPoints()
    button:SetScale(1)
    button.is_assignment = nil
    for name, frame in pairs(button.attached) do
	self:DeleteButton(frame)
    end
    button.attached = {}
    if button.assignment then
	button.assignment.attached[button:GetName()] = nil
	if not deleting then self:ArrangeChilds(button.assignment) end
    end
    button.assignment = nil
end



function DrawingBoard:DeleteButton(button)
    self:ResetButton(button, true)
    if button.type == "player" then
	button:ClearAllPoints()
	button:SetPoint("TOPLEFT", self.frames["DrawingBoardMainFrame"], "TOPLEFT", 705, -20*(button.number-1))
    end
    if button.type == "custom" then
	button:Hide()
	self.activeboard.frames[button:GetName()] = nil
    end
    if button.type == "rt" then
	button:ClearAllPoints()
	button:SetPoint("TOPLEFT", self.frames["DrawingBoardMainFrame"], "BOTTOMLEFT", (button.number-1)*(button:GetWidth()+2), -2)
    end 
end


function DrawingBoard:DropButton(button)
    local self = DrawingBoard
    if self:CheckCollision(self.frames["DrawingBoardMainFrame"], button) then
	local assignment = nil
	for name, frame in pairs(self.activeboard.frames) do
	    if frame.is_assignment == true and frame ~= button then
		if self:CheckCollision(frame, button) then
		    assignment = frame
		end
	    end
	end
	if assignment then
	    self:AttachButton(button, assignment)
	elseif not button.is_assignment then
	    local p1, parent, p2, x, y = button:GetPoint()
	    self:ResetButton(button)
	    button:SetScale(1.5)
	    button:SetPoint(p1, parent,p2,x/button:GetScale(), y/button:GetScale())
	    button.is_assignment = true
	    
	end
    else
	self:DeleteButton(button)
    end
end


function DrawingBoard:AttachButton(button, assignment)
    self:ResetButton(button)
    assignment.attached[button:GetName()] = button
    button.assignment = assignment
    
    self:ArrangeChilds(assignment)
end


function DrawingBoard:ArrangeChilds(button)
    -- Calculate new positions
    local n=0
    for name, frame in pairs(button.attached) do
	n = n + 1
    end
    
    local r = 20
    repeat
	local i = 0
	for name, frame in pairs(button.attached) do
	    frame:ClearAllPoints()
	    frame:SetPoint("CENTER", button, "CENTER", math.sin(2*math.pi*i/n)*r, math.cos(2*math.pi*i/n)*r)
	    i = i + 1
	end
	r = r + 5
    until self:ButtonsPretty(button) or r == 1000
end


function DrawingBoard:ButtonsPretty(assignment)
    for name, frame in pairs(assignment.attached) do
	if self:CheckCollision(assignment, frame) then
	    return false
	end
    end
    
    for name1, frame1 in pairs(assignment.attached) do
	for name2, frame2 in pairs(assignment.attached) do
	    if frame1 ~= frame2 and self:CheckCollision(frame1, frame2) then
	       return false
	    end
	end	
    end
 
    return true
end

-- ********* Helpers

function DrawingBoard:Debug(...)
    if self.debug then self:Print(...) end
end


function DrawingBoard:CheckCollision(button1, button2)
    local scale1 = button1:GetScale()
    local scale2 = button2:GetScale()
    
    local left1, bottom1, width1, height1 = button1:GetRect()
    left1, bottom1, width1, height1 = left1*scale1, bottom1*scale1, width1*scale1, height1*scale1
    local left2, bottom2, width2, height2 = button2:GetRect()
    left2, bottom2, width2, height2 = left2*scale2, bottom2*scale2, width2*scale2, height2*scale2
    
    if left1 < left2+width2 and bottom1+height1 > bottom2 and left1+width1 > left2 and bottom1 < bottom2+height2 then
	return true
    else
	return false
    end
end

function DrawingBoard:SendMessage(text, channel, target)
    if channel == "" or channel == nil then
	self:Print(text);
    elseif tonumber(channel) then
	ChatThrottleLib:SendChatMessage("NORMAL", "DrawingBoard", text, "CHANNEL", nil, tonumber(channel))
    elseif channel == "WHISPER" then
	ChatThrottleLib:SendChatMessage("NORMAL", "DrawingBoard", text, channel,nil, target)
    else
	ChatThrottleLib:SendChatMessage("NORMAL", "DrawingBoard", text, channel)
    end
end
